/*
 This API provides basic support for JSON format.
 Please note that:
 1. We do not support floating-point numbers.
 2. We can decode decimal, octal (0...) and hexadecimal (0x...) integer numbers.
 3. We support only UTF-8 encoding for Unicode.
 4. We keep order of pairs within objects.

 Author: Alexandre.Khodakovski@Oracle.com
*/
#ifndef JSON_H
#define JSON_H

#ifdef __cplusplus
extern "C"
{
#endif

#ifdef _MSC_VER // Microsoft C
#ifdef OVMAPI_EXPORTS
#define JSON_API __declspec(dllexport)
#else
#define JSON_API __declspec(dllimport)
#endif
#else // not Microsoft C
#define JSON_API
#endif

#define JSON_ERROR_OUT_OF_MEMORY        -1
#define JSON_ERROR_UNEXPECTED_CHARACTER -2
#define JSON_ERROR_INVALID_VALUE        -3
#define JSON_ERROR_INVALID_NUMBER       -4

enum json_type
{
   JSON_NULL,    // null
   JSON_OBJECT,  // {"string1":value1,"string2":value2...}
   JSON_ARRAY,   // [value1,value2...]
   JSON_STRING,  // "string"
   JSON_INTEGER, // -123456789, +01234567, 0x123456789ABCDEF
   JSON_BOOLEAN  // true, false
};

typedef long json_integer;
typedef int  json_boolean;

struct json_value
{
   enum json_type type;
   union
   {
      struct json_pair *object;   // pointer to the first object pair
      struct json_element *array; // pointer to the first array element
      char *string;
      json_integer integer;
      json_boolean boolean;
   } as;
};

struct json_element // array element
{
   struct json_value value;
   struct json_element *next; // pointer to the next array element
};

struct json_pair
{
   char *name;
   struct json_value value;
   struct json_pair *next; // pointer to the next pair within object
};

/*
 Parse json string. The result is the pointer to the first pair of the root
 object or NULL if the root object is empty or if parsing failed.
 If error_position is not NULL then *error_position is set to NULL on success
 or to the position within json string where error occurred.
 If error_code is not NULL then *error_code is set to 0 on success or to one of
 (negative) JSON_ERROR... values or to the (positive) code of the character
 missing at the error position.
 Note that if you are parsing a json with an empty root object then you must
 supply a non-NULL value in error_code to distinguish between a success and
 failure.
 Use json_free to destroy the result.
*/
JSON_API struct json_pair *json_parse(const char *json, char *error_code, const char **error_position);

/*
 Find the first pair with the specified name starting from (and including) the
 specified pair. The result is the pair found or NULL. You can find subsequent
 pairs with the same name starting the search each time from the pair next to
 the last found one. NULL names are compared as empty strings ("").
*/
JSON_API struct json_pair *json_find(struct json_pair *pair, const char *name);

/*
 Print JSON object to json string and return the number of characters written
 (including the trailing 0). The json_size must be bigger than or equal to that
 number. If json string is NULL or json_size is not big enough then return the
 required size of json string as a negative value. Return 0 if JSON object
 contains an invalid value.
 NULL strings are printed as empty strings ("").
*/
JSON_API int json_print(struct json_pair *object, char *json, int json_size);

static __inline int json_size(struct json_pair *object)
{
   return -json_print(object, NULL, 0);
}

/*
 Destroy the JSON object created by json_parse or json_add... functions.
*/
JSON_API void json_free(struct json_pair *object);

/*
 Use json_add... functions to manually construct a JSON structure. Use
 json_free to destroy it. These functions add new element/pair to the end of
 the specified array/object and return pointer to what has been added or NULL
 if we've run out of memory or if the specified array/object is NULL or if the
 specified value type is wrong. All strings are copied. Array/object values
 are assigned "as is".
*/
JSON_API struct json_element *json_add_to_array(struct json_element **array, const void *value, enum json_type type);
JSON_API struct json_pair *json_add_to_object(struct json_pair **object, const char *name, const void *value, enum json_type type);

static __inline struct json_pair **json_add_object_to_array(struct json_element **array, struct json_pair *value)
{
   struct json_element *result = json_add_to_array(array, value, JSON_OBJECT);
   if (result == NULL) return NULL;
   return &(result->value.as.object);
}

static __inline struct json_pair **json_add_object_to_object(struct json_pair **object, const char *name, struct json_pair *value)
{
   struct json_pair *result = json_add_to_object(object, name, value, JSON_OBJECT);
   if (result == NULL) return NULL;
   return &(result->value.as.object);
}

static __inline struct json_element **json_add_array_to_array(struct json_element **array, struct json_element *value)
{
   struct json_element *result = json_add_to_array(array, value, JSON_ARRAY);
   if (result == NULL) return NULL;
   return &(result->value.as.array);
}

static __inline struct json_element **json_add_array_to_object(struct json_pair **object, const char *name, struct json_element *value)
{
   struct json_pair *result = json_add_to_object(object, name, value, JSON_ARRAY);
   if (result == NULL) return NULL;
   return &(result->value.as.array);
}

static __inline char **json_add_string_to_array(struct json_element **array, const char *value)
{
   struct json_element *result = json_add_to_array(array, value, JSON_STRING);
   if (result == NULL) return NULL;
   return &(result->value.as.string);
}

static __inline char **json_add_string_to_object(struct json_pair **object, const char *name, const char *value)
{
   struct json_pair *result = json_add_to_object(object, name, value, JSON_STRING);
   if (result == NULL) return NULL;
   return &(result->value.as.string);
}

static __inline json_integer *json_add_integer_to_array(struct json_element **array, json_integer value)
{
   struct json_element *result = json_add_to_array(array, &value, JSON_INTEGER);
   if (result == NULL) return NULL;
   return &(result->value.as.integer);
}

static __inline json_integer *json_add_integer_to_object(struct json_pair **object, const char *name, json_integer value)
{
   struct json_pair *result = json_add_to_object(object, name, &value, JSON_INTEGER);
   if (result == NULL) return NULL;
   return &(result->value.as.integer);
}

static __inline json_boolean *json_add_boolean_to_array(struct json_element **array, json_boolean value)
{
   struct json_element *result = json_add_to_array(array, &value, JSON_BOOLEAN);
   if (result == NULL) return NULL;
   return &(result->value.as.boolean);
}

static __inline json_boolean *json_add_boolean_to_object(struct json_pair **object, const char *name, json_boolean value)
{
   struct json_pair *result = json_add_to_object(object, name, &value, JSON_BOOLEAN);
   if (result == NULL) return NULL;
   return &(result->value.as.boolean);
}

#ifdef __cplusplus
}
#endif

#endif // JSON_H
